#cython: language_level=3
#cython: boundscheck=False
#cython: wraparound=False

from libcpp.vector cimport vector
from libc.math cimport sqrt
cimport cython

import numpy as np

import pystan.constants


# autocovariance is a template function, which Cython doesn't yet support
cdef extern from "stan/math/prim/mat/fun/autocovariance.hpp" namespace "stan::math":
    void stan_autocovariance "stan::math::autocovariance<double>"(const vector[double]& y, vector[double]& acov)

cdef extern from "stan/math/prim/mat/fun/sum.hpp" namespace "stan::math":
    double stan_sum "stan::math::sum"(vector[double]& x)

cdef extern from "stan/math/prim/mat/fun/mean.hpp" namespace "stan::math":
    double stan_mean "stan::math::mean"(vector[double]& x)

cdef extern from "stan/math/prim/mat/fun/variance.hpp" namespace "stan::math":
    double stan_variance "stan::math::variance"(vector[double]& x)


cdef void get_kept_samples(dict sim, int k, int n, vector[double]& samples):
    """

    Parameters
    ----------
    k : unsigned int
        Chain index
    n : unsigned int
        Parameter index
    """
    cdef int i
    cdef long[:] warmup2 = np.array(sim['warmup2'])

    slst = sim['samples'][k]['chains']  # chain k, an OrderedDict
    param_names = list(slst.keys())  # e.g., 'beta[1]', 'beta[2]', ...
    cdef double[:] nv = slst[param_names[n]]  # parameter n
    samples.clear()
    for i in range(nv.shape[0] - warmup2[k]):
        samples.push_back(nv[warmup2[k] + i])


cdef double get_chain_mean(dict sim, int k, int n):
    cdef long[:] warmup2 = np.array(sim['warmup2'])
    slst = sim['samples'][k]['chains']  # chain k, an OrderedDict
    param_names = list(slst.keys())  # e.g., 'beta[1]', 'beta[2]', ...
    cdef vector[double] nv = slst[param_names[n]]  # parameter n
    return stan_mean(nv[warmup2[k]:])


cdef vector[double] autocovariance(dict sim, int k, int n):
    """
    Returns the autocovariance for the specified parameter in the
    kept samples of the chain specified.

    Parameters
    ----------
    k : unsigned int
        Chain index
    n : unsigned int
        Parameter index

    Returns
    -------
    acov : vector[double]

    Note
    ----
    PyStan is profligate with memory here in comparison to RStan. A variety
    of copies are made where RStan passes around references. This is done
    mainly for convenience; the Cython code is simpler.
    """
    cdef vector[double] samples, acov
    get_kept_samples(sim, k, n, samples)
    stan_autocovariance(samples, acov)
    return acov


def effective_sample_size(dict sim, int n):
    """
    Return the effective sample size for the specified parameter
    across all kept samples.

    This implementation matches BDA3's effective size description.

    Current implementation takes the minimum number of samples
    across chains as the number of samples per chain.

    Parameters
    ----------
    sim : dict
        Contains samples as well as related information (warmup, number
        of iterations, etc).
    n : int
        Parameter index

    Returns
    -------
    ess : int
    """
    cdef int i, chain
    cdef int m = sim['chains']

    cdef vector[int] ns_save = sim['n_save']
    cdef vector[int] ns_warmup2 = sim['warmup2']
    cdef vector[int] ns_kept = [s - w for s, w in zip(sim['n_save'], sim['warmup2'])]

    cdef int n_samples = min(ns_kept)

    cdef vector[vector[double]] acov
    for chain in range(m):
        acov.push_back(autocovariance(sim, chain, n))

    cdef vector[double] chain_mean
    cdef vector[double] chain_var
    # double rather than int to deal with Cython quirk, see issue #186
    cdef double n_kept_samples
    for chain in range(m):
        n_kept_samples = ns_kept[chain]
        chain_mean.push_back(get_chain_mean(sim, chain, n))
        chain_var.push_back(acov[chain][0] * n_kept_samples / (n_kept_samples-1))

    cdef double mean_var = stan_mean(chain_var)
    cdef double var_plus = mean_var * (n_samples-1) / n_samples

    if m > 1:
        var_plus = var_plus + stan_variance(chain_mean)

    cdef vector[double] rho_hat_t
    for _ in range(n_samples):
        rho_hat_t.push_back(0)
    cdef double rho_hat = 0
    cdef int max_t = 0
    cdef int t = 1
    cdef vector[double] acov_t
    while t < n_samples and rho_hat >= 0:
        acov_t.clear()
        for chain in range(m):
            acov_t.push_back(acov[chain][t])
        rho_hat = 1 - (mean_var - stan_mean(acov_t)) / var_plus
        if rho_hat >= 0:
            rho_hat_t[t] = rho_hat
        max_t = t
        t += 1
    cdef double ess = m * n_samples
    if max_t > 1:
        ess = ess / (1 + 2 * stan_sum(rho_hat_t))
    return ess


@cython.cdivision(True)
def split_potential_scale_reduction(dict sim, int n):
    """
    Return the split potential scale reduction (split R hat) for the
    specified parameter.

    Current implementation takes the minimum number of samples
    across chains as the number of samples per chain.

    Parameters
    ----------
    n : unsigned int
        Parameter index

    Returns
    -------
    rhat : float
        Split R hat

    """
    cdef int i, chain
    cdef double srhat

    cdef int n_chains = sim['chains']

    cdef vector[int] ns_save = sim['n_save']
    cdef vector[int] ns_warmup2 = sim['warmup2']
    cdef vector[int] ns_kept = [s - w for s, w in zip(sim['n_save'], sim['warmup2'])]

    cdef int n_samples = min(ns_kept)

    if n_samples % 2 == 1:
        n_samples = n_samples - 1

    cdef vector[double] split_chain_mean, split_chain_var
    cdef vector[double] samples, split_chain
    for chain in range(n_chains):
        samples.clear()
        get_kept_samples(sim, chain, n, samples)
        split_chain.clear()
        for i in range(n_samples / 2):
            split_chain.push_back(samples[i])
        split_chain_mean.push_back(stan_mean(split_chain))
        split_chain_var.push_back(stan_variance(split_chain))

        split_chain.clear()
        for i in range(n_samples / 2, n_samples):
            split_chain.push_back(samples[i])
        split_chain_mean.push_back(stan_mean(split_chain))
        split_chain_var.push_back(stan_variance(split_chain))

    cdef double var_between = n_samples / 2 * stan_variance(split_chain_mean)
    cdef double var_within = stan_mean(split_chain_var)

    if var_within < pystan.constants.EPSILON:
        srhat = float('nan')
    else:
        srhat = sqrt((var_between/var_within + n_samples / 2 - 1)/ (n_samples / 2))
    return srhat


def _test_autocovariance(dict sim, int k, int n):
    '''Test point for autocovariance function'''
    return autocovariance(sim, k, n)


def _test_stan_functions():
    y = np.arange(10)
    cdef vector[double] acov
    stan_autocovariance(y, acov)
    assert sum(acov) == -40.0, sum(acov)
    assert stan_sum(y) == sum(y)
    assert stan_mean(y) == np.mean(y)
    assert stan_variance(y) == np.var(y, ddof=1)



import unittest

from pystan import StanModel, stan

class TestCovExpQuad(unittest.TestCase):
    """
    Unit test added to address issue discussed here: https://github.com/stan-dev/pystan/issues/271
    """
    
    covexpquad = """
        data {
        real rx1[5];
    }
    model {
        matrix[5,5] a;

        a = cov_exp_quad(rx1, 1, 1);
    }
        """

    model = StanModel(model_code=covexpquad)

    def test_8schools_pars(self):
        model = self.model
        self.assertIsNotNone(model)
 



# distutils: language = c++
#-----------------------------------------------------------------------------
# Copyright (c) 2013, Allen B. Riddell
#
# This file is licensed under Version 3.0 of the GNU General Public
# License. See LICENSE for a text of the license.
#-----------------------------------------------------------------------------
from pystan.stanc cimport PyStancResult, stanc as c_stanc

def stanc(bytes model_stancode, bytes model_name):
    cdef PyStancResult result
    c_stanc(model_stancode, model_name, result)
    return {'status': result.status,
            'msg': result.msg.decode('utf-8'),
            'model_cppname': result.model_cppname.decode('ascii'),
            'cppcode': result.cppcode.decode('ascii')}



from collections import OrderedDict
import math
import os
import unittest

import numpy as np
import pystan
import pystan.chains


class TestChains(unittest.TestCase):
    # REF: runit.test.chains.R

    @classmethod
    def setUpClass(cls):
        testdata_path = os.path.join(os.path.dirname(__file__), 'data')
        f1 = os.path.join(testdata_path, 'blocker1.csv')
        f2 = os.path.join(testdata_path, 'blocker2.csv')

        # read csv using numpy
        c1 = np.loadtxt(f1, skiprows=21, delimiter=',')[:, 2:]
        c1_colnames = open(f1, 'r').readlines()[20].strip().split(',')[2:]
        c2 = np.loadtxt(f2, skiprows=21, delimiter=',')[:, 2:]
        c2_colnames = open(f2, 'r').readlines()[20].strip().split(',')[2:]
        np.testing.assert_equal(c1_colnames, c2_colnames)

        n_samples = len(c1)

        c1 = OrderedDict((k, v) for k, v in zip(c1_colnames, c1.T))
        c2 = OrderedDict((k, v) for k, v in zip(c2_colnames, c2.T))

        cls.lst = dict(fnames_oi=c1_colnames, samples=[{'chains': c1}, {'chains': c2}],
                       n_save=np.repeat(n_samples, 2), permutation=None,
                       warmup=0, warmup2=[0, 0], chains=2, n_flatnames=len(c1))

    def test_essnrhat(self):
        lst = self.lst
        ess = pystan.chains.ess(lst, 2)
        self.assertAlmostEqual(ess, 13.0778, delta=1)
        ess2 = pystan.chains.ess(lst, 45)
        self.assertAlmostEqual(ess2, 43.0242, delta=3)

    def test_rhat(self):
        lst = self.lst
        rhat = pystan.chains.splitrhat(lst, 2)
        self.assertAlmostEqual(rhat, 1.187, delta=0.001)
        rhat2 = pystan.chains.splitrhat(lst, 45)
        self.assertAlmostEqual(rhat2, 1.03715, delta=0.001)

    def test_rhat_zero(self):
        model_code = """
          parameters { real x; }
          transformed parameters { real y; y <- 0.0; }
          model {x ~ normal(0, 1);}
          """
        model = pystan.StanModel(model_code=model_code)
        fit = model.sampling()
        rhat = pystan.chains.splitrhat(fit.sim, 1)
        self.assertTrue(math.isnan(rhat))


class TestChainsNormal(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        testdata_path = os.path.join(os.path.dirname(__file__), 'data')
        f1 = os.path.join(testdata_path, 'normal1.csv')
        f2 = os.path.join(testdata_path, 'normal2.csv')

        # samples drawn using Stan 2.11.0
        # read csv using numpy
        c1 = np.loadtxt(f1, skiprows=42, delimiter=',')[:, 2:]
        c1_colnames = open(f1, 'r').readlines()[37].strip().split(',')[2:]
        c2 = np.loadtxt(f2, skiprows=42, delimiter=',')[:, 2:]
        c2_colnames = open(f2, 'r').readlines()[37].strip().split(',')[2:]
        np.testing.assert_equal(c1_colnames, c2_colnames)

        assert len(c1) == len(c2) == 40, (len(c1), len(c2))
        y_index = c1_colnames.index('y')
        y_mean = np.concatenate([c1[:, y_index], c2[:, y_index]]).mean()
        np.testing.assert_almost_equal(y_mean, -0.05038095325)

        n_samples = len(c1)

        c1 = OrderedDict((k, v) for k, v in zip(c1_colnames, c1.T))
        c2 = OrderedDict((k, v) for k, v in zip(c2_colnames, c2.T))

        cls.lst = dict(fnames_oi=c1_colnames, samples=[{'chains': c1}, {'chains': c2}],
                       n_save=np.repeat(n_samples, 2), permutation=None,
                       warmup=0, warmup2=[0, 0], chains=2, n_flatnames=len(c1))


    def test_essnrhat(self):
        lst = self.lst
        param_index = lst['fnames_oi'].index('y')
        assert param_index == 5
        ess = pystan.chains.ess(lst, param_index)
        self.assertAlmostEqual(ess, 53.7651, delta=1)

    def test_rhat(self):
        lst = self.lst
        param_index = lst['fnames_oi'].index('y')
        rhat = pystan.chains.splitrhat(lst, param_index)
        self.assertAlmostEqual(rhat, 9.9254714e-01, delta=0.000001)
